@@ -16,6 +16,8 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+import os
+
 from gi.repository import Gdk
 from gi.repository import GLib
 from gi.repository import Gtk
@@ -33,7 +35,7 @@ class Dialog(Window):
 
     def __init__(self, widget=None, parent=None, content_box=None, buttons_start=(), buttons_end=(),
                  default_button=None, show_callback=None, close_callback=None, title="", width=0, height=0,
-                 modal=True, resizable=True, show_title=True, show_title_buttons=True):
+                 modal=True, show_title=True, show_title_buttons=True):
 
         self.parent = parent
         self.modal = modal
@@ -54,15 +56,20 @@ class Dialog(Window):
             destroy_with_parent=True,
             default_width=width,
             default_height=height,
-            resizable=resizable,
             child=container
         )
         super().__init__(widget)
         Accelerator("Escape", widget, self.close)
 
         if GTK_API_VERSION == 3:
-            widget.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)  # pylint: disable=no-member
-            widget.set_type_hint(Gdk.WindowTypeHint.DIALOG)           # pylint: disable=no-member
+            if os.environ.get("GDK_BACKEND") == "broadway":
+                # Workaround for dialogs being centered at (0,0) coords on startup
+                position = Gtk.WindowPosition.CENTER
+            else:
+                position = Gtk.WindowPosition.CENTER_ON_PARENT
+
+            widget.set_position(position)                    # pylint: disable=no-member
+            widget.set_type_hint(Gdk.WindowTypeHint.DIALOG)  # pylint: disable=no-member
 
         if content_box:
             content_box.set_vexpand(True)
@@ -166,11 +173,6 @@ class Dialog(Window):
 
         # Hide the dialog
         self.widget.set_visible(False)
-
-        # "Soft-delete" the dialog. This is necessary to prevent the dialog from
-        # appearing in window peek on Windows
-        self.widget.unrealize()
-
         return True
 
     def _set_dialog_properties(self):
@@ -182,6 +184,10 @@ class Dialog(Window):
 
         self.widget.connect("show", self._on_show)
 
+        # Make all dialogs resizable to fix positioning issue.
+        # Workaround for https://gitlab.gnome.org/GNOME/mutter/-/issues/3099
+        self.widget.set_resizable(True)
+
         if self.parent:
             self.widget.set_transient_for(self.parent.widget)
 
@@ -315,7 +321,7 @@ class MessageDialog(Window):
         )
         self.message_label = Gtk.Label(
             margin_bottom=2, halign=Gtk.Align.CENTER, label=message, valign=Gtk.Align.START, vexpand=True, wrap=True,
-            max_width_chars=60, use_markup=True, visible=True
+            max_width_chars=60, visible=True
         )
 
         add_css_class(title_label, "title-2")
@@ -358,7 +364,7 @@ class MessageDialog(Window):
             action_box.add(action_area)                               # pylint: disable=no-member
 
         for response_type, button_label in buttons:
-            button = Gtk.Button(label=button_label, use_underline=True, hexpand=True, visible=True)
+            button = Gtk.Button(use_underline=True, hexpand=True, visible=True)
             button.connect("clicked", self._on_button_pressed, response_type)
 
             if GTK_API_VERSION >= 4:
@@ -368,9 +374,8 @@ class MessageDialog(Window):
 
             if response_type == self.destructive_response_id:
                 add_css_class(button, "destructive-action")
-                continue
 
-            if response_type in {"cancel", "ok"}:
+            elif response_type in {"cancel", "ok"}:
                 if GTK_API_VERSION >= 4:
                     widget.set_default_widget(button)  # pylint: disable=no-member
                 else:
@@ -380,6 +385,10 @@ class MessageDialog(Window):
                 self.message_label.set_mnemonic_widget(button)
                 self.default_focus_widget = button
 
+            # Set mnemonic widget before button label in order for screen reader to
+            # read both labels
+            button.set_label(button_label)
+
         return widget
 
     def _add_long_message(self, text):
@@ -457,7 +466,6 @@ class OptionDialog(MessageDialog):
     def _add_option_toggle(self, option_label, option_value):
 
         toggle = Gtk.CheckButton(label=option_label, active=option_value, receives_default=True, visible=True)
-        self.message_label.set_mnemonic_widget(toggle)
 
         if option_label:
             if GTK_API_VERSION >= 4:
@@ -488,11 +496,13 @@ class EntryDialog(OptionDialog):
         ], **kwargs)
 
         self.entry_container = None
+        self.entry_combobox = None
+        self.second_entry_combobox = None
+
         self.entry_combobox = self.default_focus_widget = self._add_entry_combobox(
             default, activates_default=not use_second_entry, visibility=visibility,
             show_emoji_icon=show_emoji_icon, droplist=droplist
         )
-        self.second_entry = None
 
         if use_second_entry:
             self.second_entry_combobox = self._add_entry_combobox(
@@ -518,8 +528,8 @@ class EntryDialog(OptionDialog):
 
             combobox.unfreeze()
 
-        if activates_default:
-            self.message_label.set_mnemonic_widget(entry if activates_default else combobox.widget)
+        if self.entry_combobox is None:
+            self.message_label.set_mnemonic_widget(entry if has_entry else combobox.widget)
 
         self.container.set_visible(True)
         return combobox
@@ -539,7 +549,7 @@ class EntryDialog(OptionDialog):
         else:
             self.entry_container.add(entry)     # pylint: disable=no-member
 
-        if activates_default:
+        if self.entry_combobox is None:
             self.message_label.set_mnemonic_widget(entry)
 
         self.container.set_visible(True)
