@@ -25,7 +25,6 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
-import os.path
 import time
 
 from pynicotine import slskmessages
@@ -161,15 +160,15 @@ class Uploads(Transfers):
     # Stats/Limits #
 
     @staticmethod
-    def _get_file_size(file_path):
+    def _get_current_file_size(file_path):
 
         try:
-            size = os.path.getsize(encode_path(file_path))
-        except Exception:
-            # file doesn't exist (remote files are always this)
-            size = 0
+            new_size = os.path.getsize(encode_path(file_path))
 
-        return size
+        except Exception:
+            new_size = None
+
+        return new_size
 
     def get_downloading_users(self):
         return set(self.active_users).union(self.queued_users)
@@ -233,11 +232,14 @@ class Uploads(Transfers):
 
         return self.total_bandwidth >= bandwidth_limit
 
-    def is_new_upload_accepted(self):
+    def is_new_upload_accepted(self, enforce_limits=True):
 
         if core.shares is None or core.shares.rescanning:
             return False
 
+        if not enforce_limits:
+            return True
+
         if config.sections["transfers"]["useupslots"]:
             # Limit by upload slots
             if self.is_slot_limit_reached():
@@ -440,7 +442,21 @@ class Uploads(Transfers):
     def _clear_transfer(self, transfer, denied_message=None, update_parent=True):
 
         self._abort_transfer(transfer, denied_message=denied_message)
-        del self.transfers[transfer.username + transfer.virtual_path]
+
+        log.add_transfer("Clearing upload %(path)s to user %(user)s", {
+            "path": transfer.virtual_path,
+            "user": transfer.username
+        })
+
+        try:
+            del self.transfers[transfer.username + transfer.virtual_path]
+
+        except KeyError:
+            log.add(("FIXME: failed to remove upload %(path)s to user %(user)s, not "
+                     "present in list"), {
+                "path": transfer.virtual_path,
+                "user": transfer.username
+            })
 
         events.emit("clear-upload", transfer, update_parent)
 
@@ -460,6 +476,7 @@ class Uploads(Transfers):
         # Is user allowed to download?
         ip_address, _port = addr
         permission_level, reject_reason = core.shares.check_user_permission(username, ip_address)
+        size = None
 
         if permission_level == PermissionLevel.BANNED:
             reject_message = TransferRejectReason.BANNED
@@ -467,19 +484,19 @@ class Uploads(Transfers):
             if reject_reason:
                 reject_message += f" ({reject_reason})"
 
-            return False, reject_message
+            return False, reject_message, size
 
         if core.shares.rescanning:
             self._pending_network_msgs.append(msg)
-            return False, None
+            return False, None, size
 
         # Is that file already in the queue?
         if self.is_upload_queued(username, virtual_path):
-            return False, TransferRejectReason.QUEUED
+            return False, TransferRejectReason.QUEUED, size
 
         # Are we waiting for existing uploads to finish?
         if self.pending_shutdown:
-            return False, TransferRejectReason.PENDING_SHUTDOWN
+            return False, TransferRejectReason.PENDING_SHUTDOWN, size
 
         # Has user hit queue limit?
         enable_limits = True
@@ -492,16 +509,15 @@ class Uploads(Transfers):
             limit_reached, reason = self.is_queue_limit_reached(username)
 
             if limit_reached:
-                return False, reason
+                return False, reason, size
+
+        is_file_shared, size = core.shares.file_is_shared(username, virtual_path, real_path)
 
         # Do we actually share that file with the world?
-        if not core.shares.file_is_shared(username, virtual_path, real_path):
-            return False, TransferRejectReason.FILE_NOT_SHARED
+        if not is_file_shared:
+            return False, TransferRejectReason.FILE_NOT_SHARED, size
 
-        if not self.is_file_readable(virtual_path, real_path):
-            return False, TransferRejectReason.FILE_READ_ERROR
-
-        return True, None
+        return True, None, size
 
     def _get_upload_candidate(self):
         """Retrieve a suitable queued transfer for uploading.
@@ -562,32 +578,61 @@ class Uploads(Transfers):
             self._user_update_counter += 1
             self._user_update_counters[username] = self._user_update_counter
 
-    def _check_upload_queue(self):
+    def _check_upload_queue(self, upload_candidate=None):
         """Find next file to upload."""
 
-        if not self.is_new_upload_accepted():
-            return
+        final_upload_candidate = None
 
-        upload_candidate, has_active_uploads = self._get_upload_candidate()
-
-        if upload_candidate is None:
-            if not has_active_uploads and self.pending_shutdown:
-                self.pending_shutdown = False
-                core.quit()
-            return
-
-        username = upload_candidate.username
-
-        if slskmessages.UserStatus.OFFLINE in (core.users.login_status, core.users.statuses.get(username)):
-            # Either we are offline or the user we want to upload to is
-            if self._auto_clear_transfer(upload_candidate):
+        while final_upload_candidate is None:
+            # If a candidate is provided, we want to upload it immediately
+            if not self.is_new_upload_accepted(enforce_limits=(upload_candidate is None)):
                 return
 
-            self._abort_transfer(upload_candidate, status=TransferStatus.USER_LOGGED_OFF)
-            return
+            if upload_candidate is None:
+                upload_candidate, has_active_uploads = self._get_upload_candidate()
+
+                if upload_candidate is None:
+                    if not has_active_uploads and self.pending_shutdown:
+                        self.pending_shutdown = False
+                        core.quit()
+                    return
+
+            elif upload_candidate not in self.queued_users.get(upload_candidate.username, {}).values():
+                return
+
+            username = upload_candidate.username
+
+            if slskmessages.UserStatus.OFFLINE in (core.users.login_status, core.users.statuses.get(username)):
+                # Either we are offline or the user we want to upload to is
+                if self._auto_clear_transfer(upload_candidate):
+                    continue
+
+                self._abort_transfer(upload_candidate, status=TransferStatus.USER_LOGGED_OFF)
+                continue
+
+            virtual_path = upload_candidate.virtual_path
+            real_path = core.shares.virtual2real(virtual_path)
+            is_file_shared, _new_size = core.shares.file_is_shared(username, virtual_path, real_path)
+
+            if not is_file_shared:
+                self._clear_transfer(upload_candidate, denied_message=TransferRejectReason.FILE_NOT_SHARED)
+                continue
+
+            if not self.is_file_readable(virtual_path, real_path):
+                self._abort_transfer(
+                    upload_candidate, denied_message=TransferRejectReason.FILE_READ_ERROR,
+                    status=TransferStatus.LOCAL_FILE_ERROR
+                )
+                continue
+
+            current_size = self._get_current_file_size(real_path)
+
+            if current_size is not None:
+                upload_candidate.size = current_size
+
+            final_upload_candidate = upload_candidate
 
         self.token = slskmessages.increment_token(self.token)
-        virtual_path = upload_candidate.virtual_path
 
         log.add_transfer(
             "Checked upload queue, requesting to upload file %(file)s with token %(token)s to user %(user)s", {
@@ -597,16 +642,16 @@ class Uploads(Transfers):
             }
         )
 
-        self._dequeue_transfer(upload_candidate)
-        self._unfail_transfer(upload_candidate)
-        self._activate_transfer(upload_candidate, self.token)
+        self._dequeue_transfer(final_upload_candidate)
+        self._unfail_transfer(final_upload_candidate)
+        self._activate_transfer(final_upload_candidate, self.token)
 
         core.send_message_to_peer(
             username, slskmessages.TransferRequest(
                 direction=slskmessages.TransferDirection.UPLOAD, token=self.token, file=virtual_path,
-                filesize=upload_candidate.size))
+                filesize=final_upload_candidate.size))
 
-        self._update_transfer(upload_candidate)
+        self._update_transfer(final_upload_candidate)
 
     def ban_users(self, users, ban_message=None):
         """Ban a user, cancel all the user's uploads, send a 'Banned' message
@@ -630,21 +675,17 @@ class Uploads(Transfers):
 
         self._check_upload_queue()
 
-    def enqueue_upload(self, username, virtual_path, size, folder_path=None):
+    def enqueue_upload(self, username, virtual_path):
 
         transfer = self.transfers.get(username + virtual_path)
         real_path = core.shares.virtual2real(virtual_path)
-        new_size = self._get_file_size(real_path)
+        is_file_shared, size = core.shares.file_is_shared(username, virtual_path, real_path)
 
-        if new_size > 0:
-            size = new_size
+        if not is_file_shared:
+            return
 
         if transfer is None:
-            if not folder_path:
-                folder_path = os.path.dirname(real_path)
-            else:
-                folder_path = os.path.normpath(folder_path)
-
+            folder_path = os.path.dirname(real_path)
             transfer = Transfer(username, virtual_path, folder_path, size)
             self._append_transfer(transfer)
         else:
@@ -686,7 +727,7 @@ class Uploads(Transfers):
 
         if not active_uploads:
             # No active upload, transfer a queued upload immediately
-            self._check_upload_queue()
+            self._check_upload_queue(transfer)
 
     def retry_uploads(self, uploads):
         for upload in uploads:
@@ -735,6 +776,10 @@ class Uploads(Transfers):
 
         username = msg.user
 
+        if username not in core.users.watched:
+            # Skip redundant status updates from users in joined rooms
+            return
+
         if msg.status == slskmessages.UserStatus.OFFLINE:
             for upload in self.active_users.get(username, {}).copy().values():
                 if upload.status == TransferStatus.TRANSFERRING:
@@ -746,11 +791,16 @@ class Uploads(Transfers):
             for upload in self.failed_users.get(username, {}).copy().values():
                 self._abort_transfer(upload, status=TransferStatus.USER_LOGGED_OFF)
 
+            self._online_users.discard(username)
             return
 
-        for upload in self.failed_users.get(username, {}).copy().values():
-            if upload.status == TransferStatus.USER_LOGGED_OFF:
-                self._abort_transfer(upload, status=TransferStatus.CANCELLED)
+        # No need to check transfers on away status change
+        if username not in self._online_users:
+            for upload in self.failed_users.get(username, {}).copy().values():
+                if upload.status == TransferStatus.USER_LOGGED_OFF:
+                    self._abort_transfer(upload, status=TransferStatus.CANCELLED)
+
+            self._online_users.add(username)
 
     def _user_stats(self, msg):
         """Server code 36."""
@@ -818,7 +868,7 @@ class Uploads(Transfers):
         username = msg.username
         virtual_path = msg.file
         real_path = core.shares.virtual2real(virtual_path)
-        allowed, reason = self._check_queue_upload_allowed(username, msg.addr, virtual_path, real_path, msg)
+        allowed, reason, size = self._check_queue_upload_allowed(username, msg.addr, virtual_path, real_path, msg)
 
         log.add_transfer(("Upload request for file %(filename)s from user: %(user)s, "
                           "allowed: %(allowed)s, reason: %(reason)s"), {
@@ -834,7 +884,7 @@ class Uploads(Transfers):
 
             return
 
-        transfer = Transfer(username, virtual_path, os.path.dirname(real_path), self._get_file_size(real_path))
+        transfer = Transfer(username, virtual_path, os.path.dirname(real_path), size)
 
         self._append_transfer(transfer)
         self._enqueue_transfer(transfer)
@@ -886,7 +936,7 @@ class Uploads(Transfers):
 
         # Is user allowed to download?
         real_path = core.shares.virtual2real(virtual_path)
-        allowed, reason = self._check_queue_upload_allowed(username, msg.addr, virtual_path, real_path, msg)
+        allowed, reason, size = self._check_queue_upload_allowed(username, msg.addr, virtual_path, real_path, msg)
 
         if not allowed:
             if reason:
@@ -896,8 +946,7 @@ class Uploads(Transfers):
 
         # All checks passed, user can queue file!
         if not self.is_new_upload_accepted() or username in self.active_users:
-            transfer = Transfer(
-                username, virtual_path, os.path.dirname(real_path), self._get_file_size(real_path))
+            transfer = Transfer(username, virtual_path, os.path.dirname(real_path), size)
 
             self._append_transfer(transfer)
             self._enqueue_transfer(transfer)
@@ -909,7 +958,11 @@ class Uploads(Transfers):
             return slskmessages.TransferResponse(allowed=False, reason=TransferRejectReason.QUEUED, token=token)
 
         # All checks passed, starting a new upload.
-        size = self._get_file_size(real_path)
+        current_size = self._get_current_file_size(real_path)
+
+        if current_size is not None:
+            size = current_size
+
         transfer = Transfer(username, virtual_path, os.path.dirname(real_path), size)
 
         self._append_transfer(transfer)
@@ -1019,11 +1072,6 @@ class Uploads(Transfers):
 
         real_path = core.shares.virtual2real(virtual_path)
 
-        if not core.shares.file_is_shared(username, virtual_path, real_path):
-            self._abort_transfer(upload, status=TransferRejectReason.FILE_NOT_SHARED)
-            self._check_upload_queue()
-            return
-
         try:
             # Open File
             file_handle = open(encode_path(real_path), "rb")  # pylint: disable=consider-using-with
@@ -1092,7 +1140,7 @@ class Uploads(Transfers):
         if byte_difference:
             core.statistics.append_stat_value("uploaded_size", byte_difference)
 
-            if size > current_byte_offset or upload.speed is None:
+            if size > current_byte_offset or upload.speed <= 0:
                 upload.speed = int(max(0, byte_difference // max(0.1, current_time - upload.last_update)))
                 upload.time_left = (size - current_byte_offset) // upload.speed if upload.speed else 0
             else:
@@ -1118,7 +1166,7 @@ class Uploads(Transfers):
             # We finish the upload here in case the downloading peer has a slow/limited download
             # speed and finishes later than us
 
-            if upload.speed is not None:
+            if upload.speed > 0:
                 # Inform the server about the last upload speed for this transfer
                 log.add_transfer("Sending upload speed %s to the server", human_speed(upload.speed))
                 core.send_message_to_server(slskmessages.SendUploadSpeed(upload.speed))
